#import subprocess
import dpkt
import socket
import os
#import time
import struct

ETH_P_IP = dpkt.ethernet.ETH_TYPE_IP
UDP_PROTO = dpkt.ip.IP_PROTO_UDP
TCP_PROTO = dpkt.ip.IP_PROTO_TCP

STREAMING_INT = 10
STREAMING_STOP = 201

STATE_STREAMING_INIT = 0
STATE_STREAMING_GET =1
STATE_STREAMING_OK = 2
STATE_STREAMING_END = 3

TCP_ACK_FLAG = dpkt.tcp.TH_ACK
TCP_SYN_FLAG = dpkt.tcp.TH_SYN
TCP_FIN_FLAG = dpkt.tcp.TH_FIN
TCP_RST_FLAG = dpkt.tcp.TH_RST


MAXPKTSAMPLES = 51001
MAXCWIATSAMPLES = 500

MAXSTREAMINIPS = 500 # Max IPs used for streaming video content

class ConnState:
    state = None

    def __init__(self):
        self.state = STATE_STREAMING_INIT
        self.contentLength = None
        self.contentRange = None
        self.streaming = True

class ParseYouTube:
    dumpName = None
    dumpHandle = None
    pcap = None
    videoID = None
    streamingData = None

    def __init__(self, dumpName, videoID, myIP):
        print dumpName, myIP, videoID

        self.myIP = socket.inet_aton(myIP)
        self.dumpName = dumpName
        self.dumpHandle = open(self.dumpName, "r")
        self.pcap = dpkt.pcap.Reader(self.dumpHandle)
        self.videoID = videoID

        self.connStateDict = dict() # Dictionary of connection state
        # The rest are for the flow data
        self.streamingActive = -1 # number of tcp connections where state is OK
        return

    def __reopenDump(self):
        self.dumpHandle = open(self.dumpName, "r")
        self.pcap = dpkt.pcap.Reader(self.dumpHandle)

    def parseDump(self):
        for ts, frame in self.pcap:
            eth = dpkt.ethernet.Ethernet(frame)
            if type(eth) == str or eth.type != ETH_P_IP or type(eth.data) is str or type(eth.data.data) is str:
                continue
            ip = eth.data
#           ip = dpkt.ip.IP(frame[58:])
            if ip.src != self.myIP and ip.dst != self.myIP:
                continue
            if ip.p != TCP_PROTO:
                continue
            tcp = ip.data
            payload = tcp.data
            self.parseFrame(ts, ip, tcp, payload)
        return

    def parseFrame(self, ts, ip, tcp, payload):
        if ip.src == self.myIP:
            isSrc = True
            key = (ip.src, tcp.sport, ip.dst, tcp.dport)
        else:
            isSrc = False
            key = (ip.dst, tcp.dport, ip.src, tcp.sport)

        if tcp.flags & TCP_SYN_FLAG == TCP_SYN_FLAG:
            self.__handleNewConnection(isSrc, key, tcp)
            entry = self.connStateDict[key]

        if tcp.flags & TCP_FIN_FLAG  == TCP_FIN_FLAG:
            if self.connStateDict.get(key, None) is not None:
                entry = self.connStateDict[key]
                entry.end = ts
                if entry.state == STATE_STREAMING_OK:
                    self.streamingActive = self.streamingActive - 1
                entry.state = STATE_STREAMING_END                
                
        if isSrc:
            self.__handleSrc(key, ts, ip, tcp, payload)
        else:
            self.__handleDest(key, ts, ip, tcp, payload)
        return

    def __handleNewConnection(self, isSrc, key, tcp):
        entry = ConnState()
        self.connStateDict[key] = entry
        return

    def __handleSrc(self, key, ts, ip, tcp, payload):
        entry = self.connStateDict.get(key, None)
        if entry is None:
            entry = ConnState()
            self.connStateDict[key] = entry
        if entry.state == STATE_STREAMING_INIT or entry.state == STATE_STREAMING_END:
            payload = payload.lower()
            if len(payload) > 3 and payload.find("get") != -1:
                # Am I asking for videoplayback
                if payload.find("videoplayback") != -1:
                    """ if self.streamingData.get(ip.dst, None) is None:
                    Latest request is added ... true in case of ads
                    """
                    entry.state = STATE_STREAMING_GET
                    entry.start = ts
                    entry.remoteIP = ip.dst
        return

    def __handleDest(self, key, ts, ip, tcp, payload):
        entry = self.connStateDict.get(key, None)
        if entry is None:
            return
        if entry.state == STATE_STREAMING_GET:
            payload = payload.lower()
            if payload[0:4] == "http":
                if (payload.find("200 ok") != -1 or payload.find("206") != -1):
                    entry.state = STATE_STREAMING_OK
                    idx = payload.find("content-length:")
                    endIdx = idx
                    while payload[endIdx] != '\r' and payload[endIdx+1] != '\n':
                        endIdx = endIdx + 1                    
                    print payload[idx+16: endIdx+1]
                    try:
                        entry.contentLength = int(payload[idx+16: endIdx])
                    except:
                        print "ERROR"
                        entry.contentLength = 0
                    if self.streamingActive == -1:
                        self.streamingStartTime = ts
                        self.streamingActive = 0
                    self.streamingActive = self.streamingActive + 1
                    entry.streaming = True # Connection may be used for other activity
                else:
                    entry.state == STATE_STREAMING_INIT                    
            ### TODO check for content length till the content length is not found for this IP address        
        return

    def dumpData(self, suffix, duration):        
        prefix="plotData/"
        streamingCnt = 0
        streamingIPs = list()

        #### FIRST THE META
        maxLen = -1;
        for key,entry in self.connStateDict.items():
            if entry.streaming is True:
                streamingCnt = streamingCnt + 1
                if entry.contentLength > maxLen:
                    maxLen = entry.contentLength                    
            # end if
        # end for
        fh = open('contentLengths','a')
        fh.write(str(self.videoID) + " " + str(maxLen) + " " + str(duration) + " " + str(int(maxLen)/(int(duration)+0.0)) + " " + "\n")
        fh.close()
        print str(self.videoID) + " " + str(maxLen)
        return

def getDumpFileList(baseDir):
    """
    return a list containing the tuple
    [(dumpFileName, myIP, remoteIP)...]
    """
    flist = os.listdir(baseDir)
    dct = dict()
#    print lst
    for line in flist:
        if line.find("youtube-") != -1:
            ## TODO:: depends on the  file name
            print line
            print line[17:28], line[29:]
            dct[line[17:28]] = ((str(baseDir)+str(line), line[17:28], line[29:]))
    return dct

def getDurationDict():
    fp = open('masterMetaInfo.txt','r');
    dct = dict()
    for line in fp:
        elems = line.split()
        dct[elems[0]] = elems[1]
    return dct
    
def startParsing(baseDir, suffix):
    fp = open('plotData/parseMeta','r')
    vidLst = list()
    for line in fp:
        elems = line.split()
        vidLst.append(elems[0])        
    fDict = getDumpFileList(baseDir)
    durDct = getDurationDict()
    for entry in vidLst:
        elem = fDict[entry]        
        k = ParseYouTube(elem[0], elem[1], elem[2])
        k.parseDump()
        k.dumpData(suffix, durDct.get(elem[1], -1))

if __name__ == "__main__":
    startParsing("../", "")
