#import subprocess
import dpkt
import socket
import os
#import time
import struct

ETH_P_IP = dpkt.ethernet.ETH_TYPE_IP
UDP_PROTO = dpkt.ip.IP_PROTO_UDP
TCP_PROTO = dpkt.ip.IP_PROTO_TCP

STREAMING_INT = 10
STREAMING_STOP = 181

STATE_STREAMING_INIT = 0
STATE_STREAMING_GET =1
STATE_STREAMING_OK = 2
STATE_STREAMING_END = 3

TCP_ACK_FLAG = dpkt.tcp.TH_ACK
TCP_SYN_FLAG = dpkt.tcp.TH_SYN
TCP_FIN_FLAG = dpkt.tcp.TH_FIN
TCP_RST_FLAG = dpkt.tcp.TH_RST


MAXPKTSAMPLES = 51001
MAXCWIATSAMPLES = 500

MAXSTREAMINIPS = 500 # Max IPs used for streaming video content

class ConnState:
    start = None
    end = None
    state = None
    wndScale = None
    windowLst = None
    numBytes = None

    def __init__(self):
        self.start = None
        self.state = STATE_STREAMING_INIT
        self.sStartTime = None
        self.end = -1
        self.wndScale = 1
        self.streaming = False # used for streaming or not
        self.windowLst = list()
        self.remoteIP = None
        self.numBytes = 0


class ParseYouTube:
    dumpName = None
    dumpHandle = None
    pcap = None
    videoID = None
    streamingData = None

    def __init__(self, dumpName, videoID, myIP):
        print dumpName, myIP, videoID

        self.myIP = socket.inet_aton(myIP)
        self.dumpName = dumpName
        self.dumpHandle = open(self.dumpName, "r")
        self.pcap = dpkt.pcap.Reader(self.dumpHandle)
        self.videoID = videoID

        self.connStateDict = dict() # Dictionary of connection state
        # The rest are for the flow data

        self.streamingStartTime = None
        self.prevPktTime = None
        self.numPkts = 0
        self.cycleCnt = 0
        self.cycleBytes = 0
        
        self.downEvol = list()
        self.seqEvol = list()
        self.rcvWndEvol = list()        
        self.maxSeqNum = -1
        
        self.iatData = list()
        self.pktLenData = list()
        self.seqNumData = list()
        self.rcvWndData = list()
        
        self.streamingActive = -1 # number of tcp connections where state is OK
        return

    def __reopenDump(self):
        self.dumpHandle = open(self.dumpName, "r")
        self.pcap = dpkt.pcap.Reader(self.dumpHandle)

    def parseDump(self):
        for ts, frame in self.pcap:
            eth = dpkt.ethernet.Ethernet(frame)
            if type(eth) == str or eth.type != ETH_P_IP or type(eth.data) is str or type(eth.data.data) is str:
                continue
            ip = eth.data
#           ip = dpkt.ip.IP(frame[58:])
            if ip.src != self.myIP and ip.dst != self.myIP:
                continue
            if ip.p != TCP_PROTO:
                continue
            tcp = ip.data
            payload = tcp.data
            self.parseFrame(ts, ip, tcp, payload)
        return

    def parseFrame(self, ts, ip, tcp, payload):
        if ip.src == self.myIP:
            isSrc = True
            key = (ip.src, tcp.sport, ip.dst, tcp.dport)
        else:
            isSrc = False
            key = (ip.dst, tcp.dport, ip.src, tcp.sport)

        if tcp.flags & TCP_SYN_FLAG == TCP_SYN_FLAG:
            self.__handleNewConnection(ts, isSrc, key, tcp)
            entry = self.connStateDict[key]

        if tcp.flags & TCP_FIN_FLAG  == TCP_FIN_FLAG:
            if self.connStateDict.get(key, None) is not None:
                entry = self.connStateDict[key]
                entry.end = ts
                if entry.state == STATE_STREAMING_OK:
                    self.streamingActive = self.streamingActive - 1
                entry.state = STATE_STREAMING_END                
                
        if isSrc:
            self.__handleSrc(key, ts, ip, tcp, payload)
        else:
            self.__handleDest(key, ts, ip, tcp, payload)
        return

    def __handleNewConnection(self, ts, isSrc, key, tcp):       
        entry = self.connStateDict.get(key, None)
        if entry is None:
           entry = ConnState()
           self.connStateDict[key] = entry
        if isSrc is not True:
            return
        window = tcp.win
        opts = tcp.opts
        opts = dpkt.tcp.parse_opts(opts)
        for elem in opts:
            if elem[0] == 3: # KIND 3 for scale
                strScale = elem[1]
                scale = ord(strScale)
                entry.wndScale = 1<<scale    
#                print entry.wndScale
        entry.windowLst.append((ts, window * entry.wndScale))
        return

    def __handleSrc(self, key, ts, ip, tcp, payload):
        entry = self.connStateDict.get(key, None)
        if entry is None:
            entry = ConnState()
            self.connStateDict[key] = entry
        if entry.state == STATE_STREAMING_INIT or entry.state == STATE_STREAMING_END:
            payload = payload.lower()
            if len(payload) > 3 and payload.find("get") != -1:
                # Am I asking for videoplayback
                if payload.find("videoplayback") != -1:
                    """ if self.streamingData.get(ip.dst, None) is None:
                    Latest request is added ... true in case of ads
                    """
                    entry.state = STATE_STREAMING_GET
                    entry.start = ts
                    entry.remoteIP = ip.dst
        window = tcp.win
        opts = tcp.opts
        opts = dpkt.tcp.parse_opts(opts)
        for elem in opts:
            if elem[0] == 3: # KIND 3 for scale
                strScale = elem[1]
                scale = ord(strScale)
                scale = 1<<scale
                entry.wndScale = scale
#                print entry.wndScale
        entry.windowLst.append((ts, window*entry.wndScale))
        return

    def __handleDest(self, key, ts, ip, tcp, payload):
        entry = self.connStateDict.get(key, None)
        if entry is None:
            return
        if entry.state == STATE_STREAMING_GET:
            payload = payload.lower()
            if payload[0:4] == "http":
                if (payload.find("200 ok") != -1 or payload.find("206")!= -1):
                    entry.state = STATE_STREAMING_OK
                    entry.sStartTime = ts
                    entry.streaming = True
                    # New flow
                    self.maxSeqNum = tcp.seq
                    print self.maxSeqNum
                    entry.remoteIP = ip.src
                    self.prevPktTime = ts
                    if self.streamingActive == -1:
                        self.streamingStartTime = ts
                        self.streamingActive = 0
                    self.streamingActive = self.streamingActive + 1            
                    print "Streaming", ts, ts - self.streamingStartTime
                else:
                    entry.state = STATE_STREAMING_INIT     
            ### TODO check for content length till the content length is not found for this IP address        
        if self.streamingActive < 1:
            return
        if entry.state != STATE_STREAMING_OK:
            return
        ### Assumption is only one flow sends data at a time
        ### need to check if streamingActive == 1 but may not be case
        ### due to delayed FIN i.e. FIN of one flow after HTTP OK
        payloadLen = ip.len - (ip._get_hl()*4) - (tcp.off*4)
        entry.numBytes = entry.numBytes + payloadLen
        
        self.iatData.append(self.prevPktTime - ts)
        self.prevPktTime = ts
        self.pktLenData.append(payloadLen)
        self.seqNumData.append(tcp.seq)
        if len(entry.windowLst) > 0 :
            self.rcvWndData.append(entry.windowLst[-1][0])
        else:
            self.rcvWndData.append(0)
        
        if tcp.seq >= self.maxSeqNum:
            self.maxSeqNum = tcp.seq           
                
        relTs = (ts - self.streamingStartTime)*1000 #  relative time from start in ms
        if (relTs > (self.cycleCnt+1)*STREAMING_INT):
            while relTs > (self.cycleCnt+1)*STREAMING_INT:
                self.downEvol.append(self.cycleBytes)
                self.cycleBytes = 0
                #If this is made 0 we get for a cycle; we now get cumulative
                
                if len(self.seqEvol) > 0:
                    lastSeq = self.seqEvol[-1]
                else:
                    lastSeq = self.maxSeqNum
                self.seqEvol.append(lastSeq)
                sampleTs = self.streamingStartTime + ((self.cycleCnt-1)*STREAMING_INT/(1000.0))
                wndLst = entry.windowLst
                delCnt = 0
                window = 0 
                for elem in wndLst:
                    delCnt = delCnt + 1
                    if elem[0] <= sampleTs:
                        window = elem[1]
                        continue
                    else:
                        break
                delCnt = delCnt - 2
                wndLst = wndLst[delCnt:]
                entry.windowLst = wndLst                    
                self.rcvWndEvol.append(window)    
                self.cycleCnt = self.cycleCnt + 1
            # End of while loop
            self.seqEvol[-1] = self.maxSeqNum # replace the last entry with the current max
        # End of for loop                
        self.cycleBytes = self.cycleBytes + payloadLen
        return

    def dumpData(self, suffix):        
        prefix="plotData/"
        streamingCnt = 0
        streamingIPs = list()

        #### FIRST THE META
        for key,entry in self.connStateDict.items():
            if entry.streaming is True:
                streamingCnt = streamingCnt + 1
                streamingIPs.append((entry.remoteIP, entry.sStartTime, entry.end, entry.numBytes))
                print socket.inet_ntoa(entry.remoteIP), entry.sStartTime, entry.numBytes
            # end if
        # end for
                
        fh = open(str(prefix)+"parseMeta"+str(suffix),"a")                
        fh.write(str(self.videoID)+ " " + str(socket.inet_ntoa(self.myIP)) + " " + str(streamingCnt))
        fh.write("\n")
        fh.close()

        ### THE START AND END TIME OF STREAMING IPs
        fh = open(str(prefix)+"streamingIPs"+str(suffix),"ab")
        lstLen = len(streamingIPs)
        elem = struct.pack('ff', MAXSTREAMINIPS, len(streamingIPs))
        fh.write(elem)
        for j in streamingIPs:
            elem = struct.pack('fff', j[1], j[2], j[3])
            fh.write(elem)
        padding = [-1]*(MAXSTREAMINIPS - lstLen)*3
        elem = struct.pack(len(padding)*'f', *(j for j in padding))
        fh.write(elem)
        # end for
            
            
        ### Then the evol lists
        for lst, fName in [(self.iatData, str(prefix) + "parsedIAT"+str(suffix)),
                           (self.pktLenData,str(prefix) + "parsedPayloadLens"+str(suffix)),
                           (self.seqNumData, str(prefix) + "parsedSeqNums"+str(suffix))]:
            fh = open(fName, "ab");            
            lstLen = len(lst)
            elem = struct.pack('ff', MAXPKTSAMPLES,lstLen)
            fh.write(elem)
            # In case of IAT, first elem will be 0 iat between first packet and itself is 0
            if lstLen >= MAXPKTSAMPLES:
                elem = struct.pack(MAXPKTSAMPLES*'f', *(j for j in lst[0:MAXPKTSAMPLES]))
                fh.write(elem)
            else:
                elem = struct.pack(lstLen*'f', *(j for j in lst))   
                fh.write(elem)
                padding = [-1]*(MAXPKTSAMPLES - lstLen)
                elem = struct.pack(len(padding)*'f', *(j for j in padding))
                fh.write(elem)             
            fh.close()
        # End of for loop

        numSamples = (STREAMING_STOP*1000/STREAMING_INT) # 40 seconds
        for lst, fName in [(self.seqEvol, str(prefix) + "seqEvolution"+str(suffix)),
                           (self.rcvWndEvol, str(prefix) + "rcvwndEvolution"+str(suffix)),
                           (self.downEvol, str(prefix) +"downEvolution"+str(suffix))]:
            fh = open(fName, "ab");            
            lstLen = len(lst)
            elem = struct.pack('ff', numSamples, lstLen)
            fh.write(elem)
            # In case of IAT, first elem will be 0 iat between first packet and itself is 0
            if lstLen >= numSamples  :
                elem = struct.pack(numSamples*'f', *(lst[0:numSamples]))
                fh.write(elem)
            else:
                elem = struct.pack(lstLen*'f', *(lst))   
                fh.write(elem)
                padding = [0]*(numSamples - lstLen)
                elem = struct.pack(len(padding)*'f', *(padding))
                fh.write(elem)             
            fh.close()
        # end for loop
        return

def getDumpFileList(baseDir):
    """
    return a list containing the tuple
    [(dumpFileName, myIP, remoteIP)...]
    """
    flist = os.listdir(baseDir)
    lst = list()
#    print lst
    for line in flist:
        if line.find("youtube-") != -1:
            ## TODO:: depends on the  file name
            print line
            print line[17:28], line[29:]
#            sys.exit(-1)
            lst.append((str(baseDir)+str(line), line[17:28], line[29:]))
    return lst

def startParsing(baseDir, suffix):
    fList = getDumpFileList(baseDir)
    fList.sort()
    print fList
    cnt = 1
    for elem in fList:
        k = ParseYouTube(elem[0], elem[1], elem[2])
        k.parseDump()
        found = False
        for key, entry in k.connStateDict.items():
#            print socket.inet_ntoa(key[0]), socket.inet_ntoa(key[2]), entry.streaming
            if entry.streaming is not False:
                print cnt, "Streaming Is Found in Dump"
                k.dumpData(suffix)
                found = True
                cnt = cnt + 1
                break
        if found is False:
            print "Streaming not found"        
#        if cnt == 10:
#            break

if __name__ == "__main__":
    startParsing("../", "")
